# -*- coding: utf-8 -*-

##
# Tiny ai helper
# Copyright (с) Ildar Bikmamatov 2022 - 2023 <support@bayrell.org>
# License: MIT
##

import torch, time, math
import torch.multiprocessing as mp
from .utils import batch_to
  

class MultiProcessPredict():
    
    def __init__(self, dataset, model,
        batch_size=4, num_workers=None,
        predict=None, predict_obj=None
    ):
        
        if num_workers is None:
            num_workers = mp.cpu_count()
        
        self.dataset = dataset
        self.model = model
        self.batch_size = batch_size
        self.predict = predict
        self.predict_obj = predict_obj
        self.loader = None
        self.workers = []
        self.num_workers = num_workers
        self.pos = 0
    
    
    def init(self):
        
        """
        Init
        """
        
        # Init model to eval
        self.model.module.share_memory()
        self.model.module.eval()
        
        # Init dataset
        self.loader = torch.utils.data.DataLoader(
            self.dataset,
            batch_size=self.batch_size,
            drop_last=False,
            shuffle=False
        )
        
        # Init vars
        self.finish_queue = mp.Queue()
        self.worker_queue = mp.Queue()
        self.loader_queue = mp.Queue( self.num_workers * 4 )
        
        # Setup obj
        obj={
            "module": self.model.module,
            "device": self.model.device,
            "loader": self.loader,
            "predict": self.predict,
            "predict_obj": self.predict_obj,
            "worker_queue": self.worker_queue,
            "loader_queue": self.loader_queue,
            "finish_queue": self.finish_queue,
            "transform_x": self.model.transform_x,
            "transform_y": self.model.transform_y,
        }
        
        # Add workers
        for _ in range(self.num_workers):
            self.add_worker(
                mp.Process(target=dataset_predict_worker, args=(obj,))
            )
    
    
    def add_worker(self, worker):
        
        """
        Add worker
        """
        
        self.workers.append(worker)
    
    
    def start(self):
        
        """
        Start multiprocessing predict
        """
        
        # Set work flag to start
        self.finish_queue.put(1)
        
        # Start workers
        for p in self.workers:
            p.start()
        
        # Loader loop
        for batch_x, batch_y in self.loader:
            
            if self.worker_queue.empty():
                break
            
            # Send batch to workers
            success = False
            while not success and not self.worker_queue.empty():
                try:
                    self.loader_queue.put( (batch_x, batch_y), True, 0.5 )
                    success = True
                except:
                    pass
                
        
        # Set finish flag to True
        if not self.finish_queue.empty():
            self.finish_queue.get()
        
        # Join to all workers
        for p in self.workers:
            p.join()


def dataset_predict_worker(obj):
        
    """
    One thread worker
    """
    
    import gc
    
    loader = obj["loader"]
    module = obj["module"]
    device = obj["device"]
    predict = obj["predict"]
    predict_obj = obj["predict_obj"]
    worker_queue = obj["worker_queue"]
    loader_queue = obj["loader_queue"]
    finish_queue = obj["finish_queue"]
    transform_x = obj["transform_x"]
    transform_y = obj["transform_y"]
    
    worker_queue.put(1)
    
    try:
        while not ( finish_queue.empty() and loader_queue.empty() ):
            
            batch_x = None
            batch_y = None
            
            try:
                batch_x, batch_y = loader_queue.get(True, 0.5)
            except:
                pass
            
            if batch_x is None:
                continue
            
            if transform_x:
                batch_x = transform_x(batch_x)
            
            if transform_y:
                batch_y = transform_y(batch_y)
            
            if device:
                batch_x = batch_to(batch_x, device)
            
            batch_predict = module(batch_x)
            
            if predict:
                predict(batch_x, batch_y, batch_predict, predict_obj)
            
            del batch_x, batch_y, batch_predict
            
            # Clear cache
            if torch.cuda.is_available():
                torch.cuda.empty_cache()
            
            gc.collect()
    
    except Exception as e:
        print(e)
        pass
    
    finally:
        time.sleep(1)
        worker_queue.get()


def get_features_predict(batch_x, batch_y, batch_predict, predict_obj):
        
    pipe_send = predict_obj["pipe_send"]
    
    # Send features to pipe
    sz = len(batch_x)
    for i in range(sz):
        
        s = []
        if isinstance(batch_y, list):
            s = [ batch_item[i] for batch_item in batch_y ]
        s += batch_predict[i].tolist()
        s = list(map(str,s))
        s = ",".join(s)
        
        pipe_send.send(s)


def get_features_save_file(
    worker_queue, finish_queue, pipe_recv, file_name,
    dataset_count, features_count
):
    
    # Open file to write
    h = ["label", "image_id"] + [ "f_" + str(i) for i in range(features_count) ]
    h = ",".join(h)
    file = open(file_name, "w")
    file.write(h + "\n")
    
    pos = 0
    next_pos = 0
    time_start = time.time()
    while not finish_queue.empty() or not worker_queue.empty() or pipe_recv.poll():
        
        if pipe_recv.poll():
            s = pipe_recv.recv()
            file.write(s + "\n")
            
            pos = pos + 1
            if pos > next_pos:
                next_pos = pos + 16
                t = str(round(time.time() - time_start))
                print ("\r" + str(pos) + " " +
                    str(math.floor(pos / dataset_count * 100)) + "% " + t + "s", end='')
                
                file.flush()
                
        time.sleep(0.1)
    
    t = str(round(time.time() - time_start))
    print ("\r" + str(pos) + " " +
        str(math.floor(pos / dataset_count * 100)) + "% " + t + "s", end='')
    
    # Закрыть файл для записи
    file.flush()
    file.close()
    
    print("\nOk")


def save_features_mp(
    dataset, model, file_name, features_count,
    num_workers=2, batch_size=4
):
    
    """
    Multiprocess save features to CSV
    """
    
    # Variables
    pipe_recv, pipe_send = mp.Pipe()
    
    # Create dataset
    predict = MultiProcessPredict(
        dataset=dataset,
        model=model,
        batch_size=batch_size,
        num_workers=num_workers,
        predict=get_features_predict,
        predict_obj={
            "pipe_send": pipe_send,
        }
    )
    
    # Init
    predict.init()
    
    # Add save file worker
    predict.add_worker(
        mp.Process(
            target=get_features_save_file,
            args=(
                predict.worker_queue, predict.finish_queue, pipe_recv,
                file_name, len(dataset), features_count
            )
        )
    )
    
    # Start
    predict.start()


def save_features(
    dataset, model, file_name,
    features_count, batch_size=4
):
    
    """
    One thread save features to CSV
    """
    
    # Init
    model.module.eval()
    loader = torch.utils.data.DataLoader(
        dataset,
        batch_size=batch_size,
        drop_last=False,
        shuffle=False
    )
    
    # Open file to write
    h = ["label", "image_id"] + [ "f_" + str(i) for i in range(features_count) ]
    h = ",".join(h)
    file = open(file_name, "w")
    file.write(h + "\n")
    
    pos = 0
    next_pos = 0
    dataset_count = len(dataset)
    time_start = time.time()
    device = model.device
    module = model.module
    
    for batch_x, batch_y in loader:
        
        if model.transform_x:
            batch_x = model.transform_x(batch_x)
        
        if model.transform_y:
            batch_y = model.transform_y(batch_y)
        
        # Predict batch
        if device:
            batch_x = batch_to(batch_x, device)
        
        batch_predict = module(batch_x)
        
        # Save predict to file
        sz = len(batch_x)
        for i in range(sz):
            s = []
            if isinstance(batch_y, list):
                s = [ batch_item[i] for batch_item in batch_y ]
            s += batch_predict[i].tolist()
            s = list(map(str,s))
            s = ",".join(s)
            file.write(s + "\n")
        
        file.flush()
        
        # Delete batch
        del batch_x, batch_y, batch_predict
        
        # Show progress
        pos = pos + sz
        if pos > next_pos:
            next_pos = pos + 16
            t = str(round(time.time() - time_start))
            print ("\r" + str(pos) + " " +
                str(math.floor(pos / dataset_count * 100)) + "% " + t + "s", end='')
    
    file.close()


# -*- coding: utf-8 -*-

##
# Tiny ai helper
# Copyright (с) Ildar Bikmamatov 2022 - 2023 <support@bayrell.org>
# License: MIT
##

from .Model import Model
from .Trainer import Trainer
from .mp import save_features, save_features_mp

__version__ = "0.1.7-1"

__all__ = (
    "Model",
    "Trainer",
    "DatasetPredict",
    "save_features",
    "save_features_mp",
)



# -*- coding: utf-8 -*-

##
# Tiny ai helper
# Copyright (с) Ildar Bikmamatov 2022 - 2023 <support@bayrell.org>
# License: MIT
##

import torch
import numpy as np
from typing import overload
from PIL import Image, ImageDraw
from .utils import resize_image


class InsertFirstAxis(torch.nn.Module):
    
    """
    Insert first Axis for convolution layer
    """
    
    def __call__(self, t):
        t = t[:,None,:]
        return t


class InsertLastAxis(torch.nn.Module):
    
    """
    Insert last Axis for convolution layer
    """
    
    def __call__(self, t):
        t = t[:,None]
        return t


class MoveRGBToEnd(torch.nn.Module):
        
    def __call__(self, t):
        l = len(t.shape)
        t = torch.moveaxis(t, l-3, l-1)
        return t


class MoveRGBToBegin(torch.nn.Module):
        
    def __call__(self, t):
        l = len(t.shape)
        t = torch.moveaxis(t, l-1, l-3)
        return t


class ToIntImage(torch.nn.Module):
    
    def __call__(self, t):
        
        t = t * 255
        t = t.to(torch.uint8)
        
        return t


class ToFloatImage(torch.nn.Module):
    
    def __call__(self, t):
        
        t = t.to(torch.float)
        t = t / 255.0
        
        return t


class ToFloat(torch.nn.Module):
    
    def __call__(self, t):
        
        t = t.to(torch.float)
        
        return t


class ReadImage(torch.nn.Module):
    
    def __init__(self, mode=None):
        
        self.mode=mode
    
    def __call__(self, batch):
        
        res = []
        for t in batch:
        
            t = Image.open(t)
            
            if self.mode is not None and self.mode != t.mode:
                t = t.convert(self.mode)
            
            res.append(t)
        
        return res


class ImageToTensor(torch.nn.Module):
    
    def __call__(self, batch):
        
        res = torch.tensor([])
        for t in batch:
            
            t = torch.from_numpy( np.array(t) )
            t = t[None, :]
            res = torch.cat( (res, t) )
        
        return res


class ResizeImage(torch.nn.Module):
    
    def __init__(self, size, contain=True, color=None):
        
        torch.nn.Module.__init__(self)
        
        self.size = size
        self.contain = contain
        self.color = color
    
    def __call__(self, batch):
        
        res = []
        for t in batch:
            t = resize_image(t, self.size, contain=self.contain, color=self.color)
            res.append(t)
        
        return res


class NormalizeImage(torch.nn.Module):
    
    def __init__(self, mean, std, inplace=False):
        
        import torchvision
        
        torch.nn.Module.__init__(self)
        
        self.mean = mean
        self.std = std
        self.inplace = inplace
        self.normalize = torchvision.transforms.Normalize(mean=mean, std=std, inplace=inplace)
    
    def __call__(self, t):
        
        t = self.normalize(t)
        
        return t
    
    def extra_repr(self) -> str:
        return 'mean={}, std={}, inplace={}'.format(
            self.mean, self.std, self.inplace
        )


class PreparedModule(torch.nn.Module):
    
    def __init__(self, module, weight_path=None, forward=None, *args, **kwargs):
        
        torch.nn.Module.__init__(self)
        
        self.module = module
        self.weight_path = weight_path
        self._forward = forward
        
        for param in self.module.parameters():
            param.requires_grad = False
        
        self.load_weight()
    
    def forward(self, x):
        
        if self._forward:
            x = self._forward(self, x)
        else:
            x = self.module(x)
            
        return x
    
    def load_weight(self):
        """
        Load weight
        """
        if self.weight_path:
            state_dict = torch.load( self.weight_path )
            self.module.load_state_dict( state_dict )
    
    def state_dict(self, *args, destination=None, prefix='', keep_vars=False):
        pass
    
    
class Stacking(torch.nn.Module):
    
    def __init__(self, *args, is_tensor_list=True):
        torch.nn.Module.__init__(self)
        for i, module in enumerate(args):
            self.add_module(str(i), module)
        
        self.is_tensor_list = is_tensor_list
    
    def forward(self, tensor_list):
        
        device = tensor_list[0].device
        res = torch.tensor([]).to(device)
        
        keys = list(self._modules.keys())
        for index, m in enumerate(keys):
            
            if self.is_tensor_list:
                x = tensor_list[index]
            else:
                x = tensor_list
            
            if self._modules[m] is not None:
                module = self._modules[m]
                x = module(x)
            
            res = torch.cat( (res, x), dim = 1 )
            
        return res
    
    def state_dict(self, destination=None, prefix='', keep_vars=False):
        keys = self._modules.keys()
        for m in keys:
            module = self._modules[m]
            if module is not None:
                module.state_dict(
                    destination=destination,
                    prefix=prefix + m + '.',
                    keep_vars=keep_vars
                )


class Pipe():
    def __init__(self, *args):
        self.pipe = args
    
    def __call__(self, value):
        for fn in self.pipe:
            value = fn(value)
        return value


class ModuleRemoveLastClassifier(PreparedModule):
    
    def __init__(self, module, weight_path=None, *args, **kwargs):
        
        PreparedModule.__init__(self, module, weight_path, *args, **kwargs)
        
        # Remove last layer
        classifier = list(self.module.classifier.children())
        classifier = classifier[:-1]
        self.module.classifier = torch.nn.Sequential(*classifier)


class ModuleRemoveAllClassifier(PreparedModule):
    
    def __init__(self, module, weight_path=None, *args, **kwargs):
        PreparedModule.__init__(self, module, weight_path, *args, **kwargs)
    
    def forward(self, x):
        x = self.module.features(x)
        x = self.module.avgpool(x)
        return x


# -*- coding: utf-8 -*-

##
# Tiny ai helper
# Copyright (с) Ildar Bikmamatov 2022 - 2023 <support@bayrell.org>
# License: MIT
##

import torch, time, json, math, gc, os
from torch.utils.data import DataLoader, TensorDataset
from .utils import TransformDataset, list_files, \
    get_default_device, batch_to, tensor_size, load_json, summary


class Model:
    
    def __init__(self, module=None, name=None):
        self.device = None
        self.transform_x = None
        self.transform_y = None
        self.module = module
        self.optimizer = None
        self.scheduler = None
        self.loss = None
        self.acc_fn = None
        self.name = name
        self.model_path = None
        self.epoch = 0
        self.history = {}
    
    
    def set_transform_x(self, transform_x):
        self.transform_x = transform_x
        return self
    
    
    def set_transform_y(self, transform_y):
        self.transform_y = transform_y
        return self
    
    
    def set_module(self, module):
        self.module = module
        return self
    
    
    def set_optimizer(self, optimizer):
        self.optimizer = optimizer
        return self
    
    
    def set_loss(self, loss):
        self.loss = loss
        return self
    
    
    def set_scheduler(self, scheduler):
        self.scheduler = scheduler
        return self
    
    
    def set_acc(self, acc):
        self.acc_fn = acc
        return self
    
    
    def set_name(self, name):
        self.name = name
        return self
    
    
    def set_path(self, model_path):
        self.model_path = model_path
        return self
    
    
    def init(self, acc=None, optimizer=None, loss=None, scheduler=None,
        lr=1e-3, transform_x=None, transform_y=None):
        
        """
        Init model
        """
        
        if acc is not None:
            self.acc_fn = acc
        
        if transform_x is not None:
            self.transform_x = transform_x
        
        if transform_y is not None:
            self.transform_y = transform_y
        
        if loss is not None:
            self.loss = loss
        
        if optimizer is not None:
            self.optimizer = optimizer
        
        if scheduler is not None:
            self.scheduler = scheduler
        
        if self.loss is None:
            self.loss = torch.nn.MSELoss()
        
        if self.optimizer is None:
            try:
                self.optimizer = torch.optim.Adam(self.module.parameters(), lr=lr)
            except:
                pass
        
        if self.scheduler is None and self.optimizer is not None:
            self.scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau( self.optimizer )
        
    
    def to(self, device):
        self.module = self.module.to(device)
        self.device = device
    
    
    def to_gpu(self):
        self.to( get_default_device() )
    
    
    def to_cpu(self):
        self.to( torch.device("cpu") )
    
    
    def train(self):
        self.module.train()
    
    
    def eval(self):
        self.module.eval()
    
    
    def load_file(self, file_path):
        
        """
        Load model from file
        """
        
        save_metrics = torch.load(file_path)
        self.epoch = save_metrics["epoch"]
        
        # Load history
        if "history" in save_metrics:
            self.history = save_metrics["history"].copy()
        
        # Load module
        if "module" in save_metrics:
            state_dict = save_metrics["module"]
            self.module.load_state_dict(state_dict, strict=False)
        
        # Load optimizer
        if "optimizer" in save_metrics:
            state_dict = save_metrics["optimizer"]
            self.optimizer.load_state_dict(state_dict)
        
        # Load scheduler
        if "scheduler" in save_metrics:
            state_dict = save_metrics["scheduler"]
            self.scheduler.load_state_dict(state_dict)
        
        # Load loss
        if "loss" in save_metrics:
            state_dict = save_metrics["loss"]
            self.loss.load_state_dict(state_dict)
        
    
    def load(self, file_name):
        
        """
        Load model by file name
        """
        
        file_path = os.path.join(self.model_path, file_name)
        self.load_file(file_path)
    
    
    def load_epoch(self, epoch):
        
        """
        Load current epoch
        """
        
        file_path = os.path.join(self.model_path, "model-" + str(epoch) + ".data")
        self.load_file(file_path)
        
        
    def load_last(self):
        
        """
        Load last model
        """
        
        file_name = os.path.join(self.model_path, "history.json")
        
        if not os.path.exists(file_name):
            return
        
        obj = load_json(file_name)
        
        if obj is not None:
            epoch = obj["epoch"]
            self.load_epoch(epoch)
        
    
    def load_best(self):
        
        """
        Load best model
        """
        
        file_name = os.path.join(self.model_path, "history.json")
        
        if not os.path.exists(file_name):
            return
        
        obj = load_json(file_name)
        
        if obj is not None:
            best_epoch = obj["best_epoch"]
            self.load_epoch(best_epoch)
        
    
    def save_epoch(self):
        
        """
        Save train status
        """
        
        # Get metrics
        save_metrics = {}
        save_metrics["name"] = self.name
        save_metrics["epoch"] = self.epoch
        save_metrics["history"] = self.history.copy()
        save_metrics["module"] = self.module.state_dict()
        save_metrics["optimizer"] = self.optimizer.state_dict()
        save_metrics["scheduler"] = self.scheduler.state_dict()
        save_metrics["loss"] = self.loss.state_dict()
        
        # Create folder
        if not os.path.isdir(self.model_path):
            os.makedirs(self.model_path)
        
        # Save model to file
        file_name = os.path.join(self.model_path, "model-" + str(self.epoch) + ".data")
        torch.save(save_metrics, file_name)
        
        # Save history to json
        self.save_history()
    
    
    def save_history(self):
        
        """
        Save history to json
        """
        
        best_epoch = self.get_the_best_epoch()
        file_name = os.path.join(self.model_path, "history.json")
        obj = {
            "epoch": self.epoch,
            "best_epoch": best_epoch,
            "history": self.history.copy(),
        }
        json_str = json.dumps(obj, indent=2)
        file = open(file_name, "w")
        file.write(json_str)
        file.close()
    
    
    def predict(self, x, batch_size=64):
        
        """
        Predict
        """
         
        if self.transform_x is not None:
            x = self.transform_x(x)
        
        if self.device:
            x = x.to( self.device )
        
        y = self.module(x)
        
        return y
    
    
    def predict_dataset(self, dataset, predict, batch_size=64, predict_obj=None):
        
        """
        Predict dataset
        """
        
        loader = DataLoader(
            dataset,
            batch_size=batch_size,
            drop_last=False,
            shuffle=False
        )
        
        self.module.eval()
        
        pos = 0
        next_pos = 0
        dataset_count = len(dataset)
        time_start = time.time()
        
        for batch_x, batch_y in loader:
            
            if self.transform_x:
                batch_x = self.transform_x(batch_x)
            
            if self.device:
                batch_x = batch_to(batch_x, self.device)
            
            batch_predict = self.module(batch_x)
            predict(batch_x, batch_y, batch_predict, predict_obj)
            
            # Show progress
            pos = pos + len(batch_x)
            if pos > next_pos:
                next_pos = pos + 16
                t = str(round(time.time() - time_start))
                print ("\r" + str(math.floor(pos / dataset_count * 100)) + "% " + t + "s", end='')
            
            del batch_x, batch_y, batch_predict
            
            # Clear cache
            if torch.cuda.is_available():
                torch.cuda.empty_cache()
            
            gc.collect()
        
        print ("\nOk")
    
    
    def get_metrics(self, metric_name):
        
        """
        Returns metrics by name
        """
        
        res = []
        epochs = list(self.history.keys())
        for index in epochs:
            
            epoch = self.history[index]
            res2 = [ index ]
            
            if isinstance(metric_name, list):
                for name in metric_name:
                    res2.append( epoch[name] if name in epoch else 0 )
            
            else:
                res2.append( epoch[metric_name] if metric_name in epoch else 0 )
            
            res.append(res2)
            
        return res
    
    
    def get_the_best_epoch(self):
        
        """
        Returns the best epoch
        """
        
        epoch_indexes = self.get_the_best_epochs_indexes(1)
        best_epoch = epoch_indexes[0] if len(epoch_indexes) > 0 else 0
        return best_epoch
    
    
    def get_the_best_epochs_indexes(self, epoch_count=5):
        
        """
        Returns best epoch indexes
        """
        
        metrics = self.get_metrics(["acc_val", "acc_rel"])
        
        def get_key(item):
            return [100 - item[1], item[2]]

        metrics.sort(key=get_key)
        
        res = []
        res_count = 0
        metrics_len = len(metrics)
        loss_val_last = 100
        for index in range(metrics_len):
            
            res.append( metrics[index] )
            
            if loss_val_last != metrics[index][1]:
                res_count = res_count + 1
            
            loss_val_last = metrics[index][1]
            
            if res_count > epoch_count:
                break
        
        res = [ res[index][0] for index in range(len(res)) ]
        
        return res
    
    
    def save_the_best_models(self, epoch_count=5):
        
        """
        Save the best models
        """
        
        def detect_type(file_name):
            
            import re
            
            file_type = ""
            epoch_index = 0
            
            result = re.match(r'^model-(?P<id>[0-9]+)\.data$', file_name)
            if result:
                return "model", int(result.group("id"))
            
            return file_type, epoch_index
        
        
        if self.epoch > 0 and epoch_count > 0 and os.path.isdir(self.model_path):
            
            epoch_indexes = self.get_the_best_epochs_indexes(epoch_count)
            epoch_indexes.append( self.epoch )
            
            files = list_files( self.model_path )
            
            for file_name in files:
                
                file_type, epoch_index = detect_type(file_name)
                if file_type in ["model"] and \
                    epoch_index > 0 and \
                    not (epoch_index in epoch_indexes):
                    
                    file_path = os.path.join( self.model_path, file_name )
                    os.unlink(file_path)
    
    
    def summary(self, x):
        
        """
        Show model summary
        """
        
        summary(self.module, x,
            device=self.device,
            model_name=self.name,
            transform_x=self.transform_x,
        )
    
    
    def draw_history_ax(self, ax, metrics=[], label=None, legend=True, convert=None):
        
        """
        Draw history to axes
        """
        
        metrics_values = self.get_metrics(metrics)
        for index, name in enumerate(metrics):
            values = [ item[index + 1] for item in metrics_values ]
            if convert:
                values = list(map(convert, values))
            ax.plot( values, label=name)
        
        if label:
            ax.set_xlabel( label )
        
        if legend:
            ax.legend()
    
    
    def draw_history(self, metrics=[]):
        
        """
        Draw history
        """
        
        import matplotlib.pyplot as plt
        
        fig, ax = plt.subplots(1, 2, figsize=(10, 4))
        self.draw_history_ax(ax[0],
            ["acc_train", "acc_val"],
            label="Accuracy",
            convert=lambda x: x * 100
        )
        self.draw_history_ax(ax[1],
            ["loss_train", "loss_val"],
            label="Loss"
        )
        plt.show()
    
    
    def show_history(self):
        
        h = list(self.history.keys())
        h.sort()
        
        for epoch in h:
            
            res = self.history[epoch]
            
            acc_train = res["acc_train"] if "acc_train" in res else 0
            acc_val = res["acc_val"] if "acc_val" in res else 0
            acc_rel = res["acc_rel"] if "acc_rel" in res else 0
            loss_train = res["loss_train"] if "loss_train" in res else 0
            loss_val = res["loss_val"] if "loss_val" in res else 0
            res_lr = res["res_lr"] if "res_lr" in res else 0
            time = res["time"] if "time" in res else 0
            
            acc_train = str(round(acc_train * 10000) / 100)
            acc_val = str(round(acc_val * 10000) / 100)
            acc_train = acc_train.ljust(5, "0")
            acc_val = acc_val.ljust(5, "0")
            acc_rel_str = str(round(acc_rel * 100) / 100).ljust(4, "0")
            loss_train = '%.3e' % loss_train
            loss_val = '%.3e' % loss_val
            res_lr_str = str(res_lr)
            
            print (f"Epoch {epoch}, " +
                f"acc: {acc_train}%, acc_val: {acc_val}%, rel: {acc_rel_str}, " +
                f"loss: {loss_train}, loss_val: {loss_val}, lr: {res_lr_str}, " +
                f"t: {time}s"
            )


#!/usr/bin/env python3
# -*- coding: utf-8 -*- 

"""
List:
https://pypi.python.org/pypi?%3Aaction=list_classifiers
"""

from setuptools import setup, find_packages
from os.path import abspath, dirname, join

setup(
	name="tiny_ai_helper",
	version="0.1.7-1",
	description="Tiny AI Helper for PyTorch",
	long_description=open(join(abspath(dirname(__file__)), 'README.md'), encoding='utf-8').read(),
	long_description_content_type='text/markdown',
	author="Ildar Bikmamatov",
	author_email="support@bayrell.org",
	license="MIT License",
	url = "https://github.com/bayrell/ai_helper",
	packages=find_packages(),
	include_package_data = True,
	classifiers=[
		'License :: OSI Approved :: MIT License',
		'Operating System :: POSIX :: Linux',
		'Operating System :: Microsoft :: Windows :: Windows 10',
		'Programming Language :: Python',
		'Programming Language :: Python :: 3',
		'Programming Language :: Python :: 3.8',
		'Topic :: Scientific/Engineering :: Artificial Intelligence',
	],
	keywords = [
		"ai helper", "pytorch"
	],
)



# -*- coding: utf-8 -*-

##
# Tiny ai helper
# Copyright (с) Ildar Bikmamatov 2022 - 2023 <support@bayrell.org>
# License: MIT
##

import torch, math, json, os, re
import numpy as np
from torch import nn
from PIL import Image, ImageDraw


class TransformDataset(torch.utils.data.Dataset):
    
    def __init__(self, dataset, transform_x=None, transform_y=None):
        self.dataset = dataset
        self.transform_x = transform_x
        self.transform_y = transform_y
        
    def __getitem__(self, index):
        
        x, y = self.dataset[index]
        
        if x is not None and self.transform_x:
            x = self.transform_x(x)
        
        if y is not None and self.transform_y:
            y = self.transform_y(y)
        
        return x, y
        
    def __len__(self):
        return len(self.dataset)


def append_tensor(res, t):
    
    """
    Append tensor
    """
    
    t = t[None, :]
    res = torch.cat( (res, t) )
    return res


def make_index(arr, file_name=None):
    
    """
    Make index from arr. Returns dict of positions values in arr
    """
    
    res = {}
    for index in range(len(arr)):
        value = arr[index]
        if file_name is not None:
            value = value[file_name]
        res[value] = index
    
    return res


def one_hot_encoder(num_class):
    
    """
    Returns one hot encoder to num class
    """
    
    def f(t):
        if not isinstance(t, torch.Tensor):
            t = torch.tensor(t)
        t = nn.functional.one_hot(t.to(torch.int64), num_class).to(torch.float32)
        return t
    
    return f


def label_encoder(labels):
    
    """
    Returns one hot encoder from label
    """
    
    labels = make_index(labels)
    
    def f(label_name):
        
        index = labels[label_name] if label_name in labels else -1
        
        if index == -1:
            return torch.zeros( len(labels) )
        
        t = torch.tensor(index)
        return nn.functional.one_hot(t.to(torch.int64), len(labels)).to(torch.float32)
    
    return f


def bag_of_words_encoder(dictionary_sz):
    
    """
    Returns bag of words encoder from dictionary indexes.
    """
    
    def f(text_index):
        
        t = torch.zeros(dictionary_sz - 1)
        for index in text_index:
            if index > 0:
                t[index - 1] = 1
        
        return t
        
    return f


def dictionary_encoder(dictionary, max_words):
    
    """
    Returns one hot encoder from text.
    In dictionary 0 pos is empty value, if does not exists in dictionary
    """
    
    def f(text_arr):
        
        t = torch.zeros(max_words).to(torch.int64)
        text_arr_sz = min(len(text_arr), max_words)
        
        pos = 0
        for i in range(text_arr_sz):
            word = text_arr[i]
            
            if word in dictionary:
                index = dictionary[word]
                t[pos] = index
                pos = pos + 1
        
        return t
    
    return f


def batch_map(f):
    
    def transform(batch_x):
        
        res = torch.tensor([])
        
        for i in range(len(batch_x)):
            x = f(batch_x[i])
            x = x[None, :]
            res = torch.cat( (res, x) )
        
        return res.to(batch_x.device)
    
    return transform


def batch_to(x, device):
    
    """
    Move batch to device
    """
    
    if isinstance(x, list) or isinstance(x, tuple):
        for i in range(len(x)):
            x[i] = x[i].to(device)
    else:
        x = x.to(device)
    
    return x


def tensor_size(t):

    """
    Returns tensor size
    """

    sz = t.element_size()
    shape = t.shape
    params = 1

    for c in shape:
        params = params * c

    size = params * sz

    return params, size


def create_dataset_indexes(dataset, file_name):
    
    """
    Load dataset indexes
    """
    
    index = load_json( file_name )
    if index is None:
        index = list(np.random.permutation( len(dataset) ))
        save_json(file_name, index, indent=None)
    
    return index


def split_dataset(dataset, k=0.2, indexes=None):
    
    """
    Split dataset for train and validation
    """
    
    sz = len(dataset)
    train_count = round(sz * (1 - k))
    val_count = sz - train_count
    
    if indexes is None:
        indexes = list(np.random.permutation(sz))
    
    from torch.utils.data import Subset
    return [Subset(dataset, indexes[0 : train_count]), Subset(dataset, indexes[train_count : ])]
    


def get_default_device():
    """
    Returns default device
    """
    
    device = torch.device('cuda') if torch.cuda.is_available() else torch.device('cpu')
    
    return device


def get_acc_class(batch_predict, batch_y):
    """
    Returns class accuracy
    """
    
    batch_y = torch.argmax(batch_y, dim=1)
    batch_predict = torch.argmax(batch_predict, dim=1)
    acc = torch.sum( torch.eq(batch_y, batch_predict) ).item()
    
    return acc


def get_acc_binary(batch_predict, batch_y):
    """
    Returns binary accuracy
    """
    
    from torcheval.metrics import BinaryAccuracy
    
    batch_predict = batch_predict.reshape(batch_predict.shape[0])
    batch_y = batch_y.reshape(batch_y.shape[0])
    
    acc = BinaryAccuracy() \
        .to(batch_predict.device) \
        .update(batch_predict, batch_y) \
        .compute().item()
    
    return round(acc * len(batch_y))


def resize_image(image, new_size, contain=True, color=None):
   
    """
    Resize image
    """
    
    w1 = image.size[0]
    h1 = image.size[1]
    w2 = new_size[0]
    h2 = new_size[1]

    k1 = w1 / h1
    k2 = w2 / h2
    w_new = 0
    h_new = 0
    
    if k1 > k2 and contain or k1 < k2 and not contain:
        h_new = round(w2 * h1 / w1)
        w_new = w2
        
    else:
        h_new = h2
        w_new = round(h2 * w1 / h1)
    
    image_new = image.resize( (w_new, h_new) )
    image_resize = resize_image_canvas(image_new, new_size)
    del image_new
    
    return image_resize
    

def resize_image_canvas(image, size, color=None):
   
    """
    Resize image canvas
    """
    
    width, height = size
    
    if color == None:
        pixels = image.load()
        color = pixels[0, 0]
        del pixels
        
    image_new = Image.new(image.mode, (width, height), color = color)
    
    position = (
        math.ceil((width - image.size[0]) / 2),
        math.ceil((height - image.size[1]) / 2),
    )
    
    image_new.paste(image, position)
    return image_new


def show_image_in_plot(image, cmap=None, is_float=False, first_channel=False):
    
    """
    Plot show image
    """
    
    if isinstance(image, str):
        image = Image.open(image)
    
    if torch.is_tensor(image):
        if first_channel == True:
            image = torch.moveaxis(image, 0, 2)
        
        if is_float:
            image = image * 255
            image = image.to(torch.uint8)
    
    import matplotlib.pyplot as plt
    
    plt.imshow(image, cmap)
    plt.show()


def list_files(path="", recursive=True):
    
    """
        Returns files in folder
    """
    
    def read_dir(path, recursive=True):
        res = []
        items = os.listdir(path)
        for item in items:
            
            item_path = os.path.join(path, item)
            
            if item_path == "." or item_path == "..":
                continue
            
            if os.path.isdir(item_path):
                if recursive:
                    res = res + read_dir(item_path, recursive)
            else:
                res.append(item_path)
            
        return res
    
    try:
        items = read_dir( path, recursive )
            
        def f(item):
            return item[len(path + "/"):]
        
        items = list( map(f, items) )
    
    except Exception:
        items = []
    
    return items


def get_sort_alphanum_key(name):
    
    """
    Returns sort alphanum key
    """
    
    arr = re.split("([0-9]+)", name)
    
    for key, value in enumerate(arr):
        try:
            value = int(value)
        except:
            pass
        arr[key] = value
    
    arr = list(filter(lambda item: item != "", arr))
    
    return arr


def alphanum_sort(files):
    
    """
    Alphanum sort
    """
    
    files.sort(key=get_sort_alphanum_key)


def list_dirs(path=""):
    
    """
        Returns dirs in folder
    """
    
    try:
        items = os.listdir(path)
    except Exception:
        items = []
    
    return items


class JSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        if isinstance(obj, np.floating):
            return float(obj)
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)


def save_json(file_name, obj, indent=2):
    
    """
    Save json to file
    """
    
    json_str = json.dumps(obj, indent=indent, cls=JSONEncoder)
    file = open(file_name, "w")
    file.write(json_str)
    file.close()


def load_json(file_name):
    
    """
    Load json from file
    """
    
    obj = None
    file = None
    
    try:
        
        file = open(file_name, "r")
        s = file.read()
        obj = json.loads(s)
        
    except Exception:
        pass
    
    finally:
        if file:
            file.close()
            file = None
    
    return obj


def summary(module, x, y=None, model_name=None, batch_transform=None, device=None):
        
        """
        Show model summary
        """
        
        hooks = []
        layers = []
        res = {
            "params_count": 0,
            "params_train_count": 0,
            "total_size": 0,
        }
        
        def forward_hook(module, input, output):
            
            output = output[0] if isinstance(output, tuple) else output
            
            class_name = module.__class__.__module__ + "." + module.__class__.__name__
            layer = {
                "name": module.__class__.__name__,
                "class_name": module.__class__.__module__ + "." + module.__class__.__name__,
                "shape": output.shape,
                "params": 0
            }
            
            # Get weight
            if hasattr(module, "weight"):
                params, size = tensor_size(module.weight)
                res["params_count"] += params
                res["total_size"] += size
                layer["params"] += params
                
                if module.weight.requires_grad:
                    res["params_train_count"] += params
            
            # Get bias
            if hasattr(module, "bias"):
                params, size = tensor_size(module.bias)
                res["params_count"] += params
                res["total_size"] += size
                layer["params"] += params
                
                if module.bias.requires_grad:
                    res["params_train_count"] += params
            
            # Add output size
            params, size = tensor_size(output)
            res["total_size"] += size
            
            # Add layer
            layers.append(layer)
                
        def add_hooks(module):
            hooks.append(module.register_forward_hook(forward_hook))
        
        # Get batch from Dataset
        if isinstance(x, torch.utils.data.Dataset):
            loader = torch.utils.data.DataLoader(
                x,
                batch_size=2,
                drop_last=False,
                shuffle=False
            )
            it = loader._get_iterator()
            
            x, y = next(it)
        
        if batch_transform is None:
            batch_transform = getattr(module, "batch_transform", None)
        
        # Trasform
        if batch_transform is not None:
            x, _ = batch_transform(x, y)
        
        # Move to device
        if device is not None:
            x = batch_to(x, device)
        
        # Add input size
        if isinstance(x, list) or isinstance(x, tuple):
            shapes = []
            for i in range(len(x)):
                params, size = tensor_size(x[i])
                shapes.append(x[i].shape)
            res["total_size"] += size
            layers.append({
                "name": "Input",
                "shape": shapes,
                "params": 0
            })
        else:
            params, size = tensor_size(x)
            res["total_size"] += size
            layers.append({
                "name": "Input",
                "shape": x.shape,
                "params": 0
            })
        
        # Module predict
        module.apply(add_hooks)
        y = module(x)
        
        # Clear cache
        if torch.cuda.is_available():
            torch.cuda.empty_cache()
        
        # Remove hooks
        for item in hooks:
            item.remove()
        
        res['total_size'] = round(res['total_size'] / 1024 / 1024 * 100) / 100
        
        # Calc info
        values = [ ["", "Layer", "Output", "Params"] ]
        for i, layer in enumerate(layers):
            shape = layer["shape"]
            shape_str = ""
            if isinstance(shape, list):
                shape = [ "(" + ", ".join(map(str,s)) + ")" for s in shape ]
                shape_str = "[" + ", ".join(shape) + "]"
            else:
                shape_str = "(" + ", ".join(map(str,shape)) + ")"
            
            values.append([i + 1, layer["name"], shape_str, layer["params"]])
        
        # Print info
        info_sizes = [2, 7, 7, 5]
        for _, value in enumerate(values):
            for i in range(4):
                sz = len(str(value[i]))
                if info_sizes[i] < sz:
                    info_sizes[i] = sz
            
        def format_row(arr, size):
            s = "{:<"+str(size[0] + 1)+"} {:>"+str(size[1] + 2)+"}" + \
                "{:>"+str(size[2] + 5)+"} {:>"+str(size[3] + 5)+"}"
            return s.format(*arr)
        
        width = info_sizes[0] + 1 + info_sizes[1] + 2 + info_sizes[2] + 5 + info_sizes[3] + 5 + 2
        print( "=" * width )
        print( format_row(["", "Layer", "Output", "Params"], info_sizes) )
        print( "-" * width )
        
        for _, value in enumerate(values):
            print( format_row(value, info_sizes) )
        
        print( "-" * width )
        if model_name is not None:
            print( f"Model name: {model_name}" )
        print( f"Total params: {res['params_count']}" )
        print( f"Trainable params: {res['params_train_count']}" )
        print( f"Total size: {res['total_size']} MiB" )
        print( "=" * width )


def fit(model, train_dataset, val_dataset,
    batch_size=64, epochs=10, device=None,
    min_lr=1e-5, reduction='mean'
):
    
    if torch.cuda.is_available():
        torch.cuda.empty_cache()
    
    module = model.module
    optimizer = model.optimizer
    scheduler = model.scheduler
    loss_fn = model.loss
    
    batch_transform = getattr(module, "batch_transform", None)
    
    train_loader = DataLoader(
        train_dataset,
        batch_size=batch_size,
        drop_last=False,
        shuffle=True
    )
    val_loader = DataLoader(
        val_dataset,
        batch_size=batch_size,
        drop_last=False,
        shuffle=False
    )
    
    module = module.to(device)
    
    print ("Start train")
    try:
        for epoch in range(epochs):
            
            time_start = time.time()
            train_count = 0
            train_loss = 0
            train_iter = 0
            val_count = 0
            val_loss = 0
            val_iter = 0
            total_count = len(train_dataset) + len(val_dataset)
            pos = 0
            
            # train mode
            module.train()
            
            for x_batch, y_batch in train_loader:
                
                # data to device
                x_batch = x_batch.to(device)
                y_batch = y_batch.to(device)
                if batch_transform:
                    x_batch, y_batch = batch_transform(x_batch, y_batch)
                
                # set parameter gradients to zero
                optimizer.zero_grad()
                
                # forward
                y_pred = module(x_batch)
                loss = loss_fn(y_pred, y_batch)
                loss.backward()
                optimizer.step()

                # add metrics
                batch_len = len(x_batch[0]) \
                    if isinstance(x_batch, tuple) or isinstance(x_batch, list) \
                    else len(x_batch)
                train_loss += loss
                train_count += batch_len
                train_iter += 1
                pos += batch_len

                del x_batch, y_batch, y_pred, loss

                if torch.cuda.is_available():
                    torch.cuda.empty_cache()
                
                print(f"\r{round(pos / total_count * 100)}%", end="")

            
            with torch.no_grad():

                # testing mode
                module.eval()
                
                for x_batch, y_batch in val_loader:
                    
                    # data to device
                    x_batch = x_batch.to(device)
                    y_batch = y_batch.to(device)
                    if batch_transform:
                        x_batch, y_batch = batch_transform(x_batch, y_batch)
                    
                    # forward
                    y_pred = module(x_batch)
                    loss = loss_fn(y_pred, y_batch)
                    
                    # add metrics
                    batch_len = len(x_batch[0]) \
                        if isinstance(x_batch, tuple) or isinstance(x_batch, list) \
                        else len(x_batch)
                    val_loss += loss
                    val_count += batch_len
                    val_iter += 1
                    pos += batch_len
                    
                    del x_batch, y_batch, y_pred, loss

                    if torch.cuda.is_available():
                        torch.cuda.empty_cache()
                    
                    print(f"\r{round(pos / total_count * 100)}%", end="")
            
            if reduction == "mean":
                train_loss = (train_loss / train_iter).item()
                val_loss = (val_loss / val_iter).item()
            
            elif reduction == "sum":
                train_loss = (train_loss / train_count).item()
                val_loss = (val_loss / val_count).item()
            
            scheduler.step(val_loss)
            
            # Current lr
            res_lr = []
            for param_group in optimizer.param_groups:
                res_lr.append( round(param_group['lr'], 7) )
            res_lr_str = str(res_lr)
            
            time_end = time.time()
            t = round(time_end - time_start)
            
            h = {
                "loss_train": train_loss,
                "loss_val": val_loss,
                "count_train": train_count,
                "count_val": val_count,
                "res_lr": res_lr,
                "time": t,
            }
            
            model.history[epoch] = h
            
            # Print result
            train_loss = "{:.9f}".format(train_loss)
            val_loss = "{:.9f}".format(val_loss)
            
            print(f'\repoch: {epoch + 1}, ' +
                'train_loss: {train_loss}, val_loss: {val_loss}, ' +
                'lr: {res_lr_str}, t: {t}s')
            
            if res_lr[0] < min_lr:
                break

        print ("Ok")

    except KeyboardInterrupt:
        
        print ("")
        print ("Stopped manually")
        print ("")

    if torch.cuda.is_available():
        torch.cuda.empty_cache()
    



# -*- coding: utf-8 -*-

##
# Tiny ai helper
# Copyright (с) Ildar Bikmamatov 2022 - 2023 <support@bayrell.org>
# License: MIT
##

import torch, time
from .utils import get_default_device, batch_to, get_acc_class, get_acc_binary
from torch.utils.data import DataLoader, TensorDataset


class Trainer:
    
    def __init__(self):
        
        self.device = None
        self.model = None
        self.epoch = 0
        self.loss_train = 1
        self.loss_val = 1
        self.acc_train = 0
        self.acc_val = 0
        self.count_train = 0
        self.count_val = 0
        self.batch_iter = 0
        self.time_start = 0
        self.time_end = 0
        self.train_loader = None
        self.val_loader = None
        self.max_best_models = 5
        self.min_epoch = 5
        self.max_epoch = 10
        self.min_loss_val = -1
        self.min_lr = 1e-5
        self.do_training = False
        
    
    def check_is_trained(self):
        
        """
        Returns True if model is trained
        """
        
        if self.epoch >= self.max_epoch:
            return True
        
        if self.count_val > 0 and \
            (self.loss_val / self.count_val) < self.min_loss_val and \
            self.epoch >= self.min_epoch:
            return True
        
        if self.model.optimizer.param_groups[0]["lr"] < self.min_lr:
            return True
        
        return False
    
    
    def on_start_train(self):
        pass
    
    
    def on_start_epoch(self):
        pass
    
    
    def on_start_batch_train(self, batch_x, batch_y):
        pass
    
    
    def on_end_batch_train(self):
        
        # Лог обучения
        acc_train = str(round(self.acc_train / self.count_train * 100))
        batch_iter_value = round(self.batch_iter / (self.len_train + self.len_val) * 100)
        print (f"\rEpoch {self.epoch}, {batch_iter_value}%, acc: {acc_train}%", end='')
    
    
    def on_start_batch_val(self, batch_x, batch_y):
        pass
    
    
    def on_end_batch_val(self):
        
        # Лог обучения
        acc_train = str(round(self.acc_train / self.count_train * 100))
        batch_iter_value = round(self.batch_iter / (self.len_train + self.len_val) * 100)
        print (f"\rEpoch {self.epoch}, {batch_iter_value}%, acc: {acc_train}%", end='')
    
    
    def on_end_epoch(self):
        
        # Получить текущий lr
        res_lr = []
        for param_group in self.model.optimizer.param_groups:
            res_lr.append(param_group['lr'])
        res_lr_str = str(res_lr)
        
        # Результат обучения
        loss_train = '%.3e' % (self.loss_train / self.batch_train)
        loss_val = '%.3e' % (self.loss_val / self.batch_val)
        
        acc_train = self.acc_train / self.count_train
        acc_val = self.acc_val / self.count_val
        acc_rel = acc_train / acc_val if acc_val > 0 else 0
        acc_train = str(round(acc_train * 10000) / 100)
        acc_val = str(round(acc_val * 10000) / 100)
        acc_train = acc_train.ljust(5, "0")
        acc_val = acc_val.ljust(5, "0")
        acc_rel_str = str(round(acc_rel * 100) / 100).ljust(4, "0")
        time = str(round(self.time_end - self.time_start))
        
        print ("\r", end='')
        print (f"Epoch {self.epoch}, " +
            f"acc: {acc_train}%, acc_val: {acc_val}%, rel: {acc_rel_str}, " +
            f"loss: {loss_train}, loss_val: {loss_val}, lr: {res_lr_str}, " +
            f"t: {time}s"
        )
        
        # Update model history
        self.model.epoch = self.epoch
        self.model.history[self.epoch] = {
            "loss_train": self.loss_train / self.batch_train,
            "loss_val": self.loss_val / self.batch_val,
            "acc_train": self.acc_train / self.count_train,
            "acc_val": self.acc_val / self.count_val,
            "acc_rel": acc_rel,
            "count_train": self.count_train,
            "count_val": self.count_val,
            "batch_iter": self.batch_iter,
            "res_lr": res_lr,
            "time": time,
        }
        
        # Save model
        self.model.save_epoch()
        self.model.save_the_best_models(epoch_count=self.max_best_models)
    
    
    def on_end_train(self):
        pass
    
    
    def stop_training(self):
        self.do_training = False
    
    
    def calc_metrics(self, kind, batch_x, batch_y, batch_predict, loss_value):
        
        """
        Calc metrics
        """
        
        get_acc_fn = self.model.acc_fn
        acc = get_acc_fn(batch_predict, batch_y)
        loss_value_item = loss_value.item()
        batch_count = len(batch_x[0]) if isinstance(batch_x, list) else len(batch_x)
        
        if kind == "train":
            self.acc_train = self.acc_train + acc
            self.loss_train = self.loss_train + loss_value_item
            self.count_train = self.count_train + batch_count
            self.batch_train = self.batch_train + 1
            self.batch_iter = self.batch_iter + batch_count
        
        elif kind == "valid":
            self.acc_val = self.acc_val + acc
            self.loss_val = self.loss_val + loss_value_item
            self.count_val = self.count_val + batch_count
            self.batch_val = self.batch_val + 1
            self.batch_iter = self.batch_iter + batch_count
    
    
    def fit(self, model, train_dataset, val_dataset, batch_size=64, epochs=10):
        
        """
        Fit model
        """
        
        self.device = model.device
        self.model = model
        self.len_train = len(train_dataset)
        self.len_val = len(val_dataset)
        self.max_epoch = epochs
        
        self.train_loader = DataLoader(
            train_dataset,
            batch_size=batch_size,
            drop_last=False,
            shuffle=True
        )
        
        self.val_loader = DataLoader(
            val_dataset,
            batch_size=batch_size,
            drop_last=False,
            shuffle=False
        )
        
        if self.device is None:
            self.device = get_default_device()
            self.model.to(self.device)
        
        module = self.model.module
        
        try:
            self.epoch = self.model.epoch
            self.do_training = True
            
            # Start train
            self.on_start_train()
            
            while self.do_training and not self.check_is_trained():
                
                self.loss_train = 0
                self.loss_val = 0
                self.acc_train = 0
                self.acc_val = 0
                self.count_train = 0
                self.count_val = 0
                self.batch_train = 0
                self.batch_val = 0
                self.batch_iter = 0
                self.epoch = self.epoch + 1
                self.time_start = time.time()
                
                self.on_start_epoch()
                module.train()
                
                # Обучение
                for batch_x, batch_y in self.train_loader:
                    
                    if self.model.transform_x:
                        batch_x = self.model.transform_x(batch_x)
                    
                    if self.model.transform_y:
                        batch_y = self.model.transform_y(batch_y)
                    
                    batch_x = batch_to(batch_x, self.device)
                    batch_y = batch_to(batch_y, self.device)
                    
                    # Start batch
                    self.on_start_batch_train(batch_x, batch_y)
                    
                    # Predict
                    batch_predict = module(batch_x)
                    loss_value = self.model.loss(batch_predict, batch_y)
                    
                    # Calc metrics
                    self.calc_metrics("train", batch_x, batch_y, batch_predict, loss_value)
                    del batch_x, batch_y, batch_predict
                    
                    # Вычислим градиент
                    self.model.optimizer.zero_grad()
                    loss_value.backward()
                    del loss_value
                    
                    # Оптимизируем
                    self.model.optimizer.step()
                    
                    # End batch
                    self.on_end_batch_train()
                  
                    # Clear cache
                    if torch.cuda.is_available():
                        torch.cuda.empty_cache()
                
                module.eval()
                
                # Вычислим ошибку на проверочном датасете
                for batch_x, batch_y in self.val_loader:
                    
                    if self.model.transform_x:
                        batch_x = self.model.transform_x(batch_x)
                    
                    if self.model.transform_y:
                        batch_y = self.model.transform_y(batch_y)
                    
                    batch_x = batch_to(batch_x, self.device)
                    batch_y = batch_to(batch_y, self.device)
                    
                    # Start batch
                    self.on_start_batch_val(batch_x, batch_y)
                    
                    # Predict
                    batch_predict = module(batch_x)
                    loss_value = self.model.loss(batch_predict, batch_y)
                    
                    # Calc metrics
                    self.calc_metrics("valid", batch_x, batch_y, batch_predict, loss_value)
                    del batch_x, batch_y, batch_predict, loss_value
                    
                    # End batch
                    self.on_end_batch_val()
                    
                    # Clear cache
                    if torch.cuda.is_available():
                        torch.cuda.empty_cache()
                    
                # Двигаем шедулер
                self.model.scheduler.step(self.loss_val)
                self.time_end = time.time()
                
                self.on_end_epoch()
            
            self.on_end_train()
            self.do_training = False
            
        except KeyboardInterrupt:
            
            print ("")
            print ("Stopped manually")
            print ("")
            
            pass
